#include "httpserver.h"
#include "../tunnelconstructor.h"
#include "crow/mustache.h"
#include "../htmldata.h"
#include "../tunnelconstructor.h"
#include "../versionnumber.h"

HttpServer::HttpServer(const std::string &address, uint16_t port) : address_(address), port_(port)
{
    app_.loglevel(crow::LogLevel::Critical);

    setup();
}

void HttpServer::run()
{
    app_.bindaddr(address_).port(port_).multithreaded().run();
}

void HttpServer::setup()
{
    CROW_CATCHALL_ROUTE(app_)
    ([](crow::response& res) {
        res.code = 303;
        res.set_header("Location", "/");
        res.end();
    });

    CROW_ROUTE(app_, "/generate/html")([&](const crow::request& req, crow::response& res){
        res.set_header("Content-Type", "text/html");

        const auto language = lang(req);

        const auto typeStr = req.url_params.get("type");

        auto constructor = initTunnelConstructorViaUrlQuery(req, language);
        if (constructor.second == false)
        {
            res.code = 400;
            res.write( error(reinterpret_cast<const char*>(constructor.first.errorString().c_str()), language, typeStr ? std::string("/main?type=") + typeStr : "/main", fullUrl(req)));
            res.end();
            return;
        }

        if (!typeStr)
        {
            res.code = 400;
            res.write( error(reinterpret_cast<const char*>(Notepad::WebUi::errorNoRequiredParams(language)), language, "/", fullUrl(req)));
            res.end();
            return;
        }

        const auto b33Str = req.url_params.get("b33");
        const auto blindedType = b33Str ? TunnelConstructor::stringToBlindedType(b33Str) : TunnelConstructor::BlindedType::none;

        const auto text = config( reinterpret_cast<const char*>(constructor.first.generate().c_str()) , language, constructor.first.isCommentsEnabled(),
                                 blindedType, TunnelConstructor::stringToTunnelType(typeStr), fullUrl(req));
        res.write(text);
        res.end();
    });

    CROW_ROUTE(app_, "/favicon.png")([&](const crow::request& req, crow::response& res){
        res.set_header("Content-Type", "image/png");
        res.add_header("Content-Length", std::to_string(sizeof(faviconBytes)));
        res.write(std::string(reinterpret_cast<const char*>(faviconBytes), sizeof(faviconBytes)));
        res.end();
    });

    CROW_ROUTE(app_, "/main")([&](const crow::request& req, crow::response& res){
        res.set_header("Content-Type", "text/html");
        const auto language = lang(req);
        const auto typeStr = req.url_params.get("type");
        if (! typeStr)
        {
            res.code = 400;
            res.write( error(reinterpret_cast<const char*>(Notepad::WebUi::error(language)), language, "/", fullUrl(req)) );
            res.end();
            return;
        }
        const auto blindedStr = req.url_params.get("b33");
        const auto tunnelType =  TunnelConstructor::stringToTunnelType(typeStr);
        const auto isClient = TunnelConstructor::isClientType(tunnelType);
        if (! blindedStr and ! isClient)
        {
            res.write( extraPrivacyServer(language, tunnelType) );
        }
        else if (! isClient)
        {
            res.write( mainServer( language, tunnelType, TunnelConstructor::stringToBlindedType(blindedStr)));
        }
        else // client
        {
            res.write( mainClient( language, tunnelType ) );
        }
        res.end();
    });

    CROW_ROUTE(app_, "/")([&](const crow::request& req, crow::response& res){
        res.set_header("Content-Type", "text/html");
        res.write( start( lang(req) ) );
        res.end();
    });
}

std::string HttpServer::start(Notepad::Lang lang)
{
    static std::map<Notepad::Lang, std::string> cache;
    auto iter = cache.find(lang);
    if (iter != cache.end())
    {
        return iter->second;
    }

    crow::mustache::context ctx;
    ctx["lang_code"] = Notepad::langToCode(lang);
    ctx["tagline"] = reinterpret_cast<const char*>(Notepad::WebUi::tagline(lang));
    ctx["select_tunnel_type"] = reinterpret_cast<const char*>(Notepad::WebUi::startSelectTunnelType(lang));
    ctx["client_tcp"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownClientTCP(lang));
    ctx["client_udp"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownClientUDP(lang));
    ctx["server_tcp"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownServerTCP(lang));
    ctx["server_udp"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownServerUDP(lang));
    ctx["server_http"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownServerHTTP(lang));
    ctx["server_irc"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownServerIRC(lang));
    ctx["socks_proxy"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownSOCKSProxy(lang));
    ctx["http_proxy"] = reinterpret_cast<const char*>(Notepad::WebUi::startDropdownHTTPProxy(lang));
    ctx["confirm"] = reinterpret_cast<const char*>(Notepad::WebUi::startConfirm(lang));
    ctx["about"] = reinterpret_cast<const char*>(Notepad::WebUi::startAbout(lang));
    ctx["version"] = VERSION;

    const auto templ = crow::mustache::compile(HTML_START);
    cache[lang] = templ.render_string(ctx);

    return cache[lang];
}

std::string HttpServer::extraPrivacyServer(Notepad::Lang lang, TunnelType type)
{
    static std::map<std::pair<Notepad::Lang, TunnelType>, std::string> cache;
    const auto cacheKey = std::pair<Notepad::Lang, TunnelType>({lang, type});
    auto iter = cache.find(cacheKey);
    if (iter != cache.end())
    {
        return iter->second;
    }

    crow::mustache::context ctx;
    ctx["lang_code"] = Notepad::langToCode(lang);
    ctx["to_start_title"] = reinterpret_cast<const char*>(Notepad::WebUi::toStartTitle(lang));
    ctx["about"] = reinterpret_cast<const char*>(Notepad::WebUi::b33AboutServer(lang, type).c_str());
    ctx["tagline"] = reinterpret_cast<const char*>(Notepad::WebUi::b33TaglineServer(lang));
    ctx["type"] = TunnelConstructor::tunnelTypeToString(type);
    ctx["th_option"] = reinterpret_cast<const char*>(Notepad::WebUi::b33ThOption(lang));
    ctx["th_value"] = reinterpret_cast<const char*>(Notepad::WebUi::b33ThInput(lang));
    ctx["none"] = reinterpret_cast<const char*>(Notepad::WebUi::b33NoneServer(lang));
    ctx["none_short"] = reinterpret_cast<const char*>(Notepad::WebUi::b33NoneShort(lang));
    ctx["simple"] = reinterpret_cast<const char*>(Notepad::WebUi::b33SimpleServer(lang));
    ctx["simple_short"] = reinterpret_cast<const char*>(Notepad::WebUi::b33SimpleShort(lang));
    ctx["password"] = reinterpret_cast<const char*>(Notepad::WebUi::b33PasswordServer(lang));
    ctx["password_short"] = reinterpret_cast<const char*>(Notepad::WebUi::b33PasswordShort(lang));
    ctx["keys"] = reinterpret_cast<const char*>(Notepad::WebUi::b33KeysServer(lang));
    ctx["keys_short"] = reinterpret_cast<const char*>(Notepad::WebUi::b33KeysShort(lang));
    ctx["go_back"] = reinterpret_cast<const char*>(Notepad::WebUi::goBack(lang));
    ctx["confirm"] = reinterpret_cast<const char*>(Notepad::WebUi::b33Confirm(lang));
    ctx["url"] = "/main?type=" + TunnelConstructor::tunnelTypeToString(type);
    ctx["version"] = VERSION;

    static const auto templ = crow::mustache::compile(HTML_B33_SERVER);
    cache[cacheKey] = templ.render_string(ctx);
    return cache[cacheKey];
}

Notepad::Lang HttpServer::lang(const crow::request &req)
{
    const auto lang = req.url_params.get("lang");
    if (lang)
    {
        auto& ctx = app_.get_context<crow::CookieParser>(req);
        ctx.set_cookie("lang", lang).path("/").max_age(60*60*24).httponly();
        return Notepad::stringToLang(lang);
    }

    auto& ctxCookie = app_.get_context<crow::CookieParser>(req);
    const std::string langFromCookie = ctxCookie.get_cookie("lang");
    if (langFromCookie.empty())
    {
        auto& ctxFromHeader = app_.get_context<LanguageHandler>(req);
        return ctxFromHeader.lang;
    }

    return Notepad::stringToLang(langFromCookie);
}

std::pair<TunnelConstructor, bool> HttpServer::initTunnelConstructorViaUrlQuery(const crow::request &req, Notepad::Lang lang)
{
    TunnelConstructor constructor;
    constructor.setLang(lang);

    bool comments = false;
    int8_t inbound_length = 3;
    int8_t outbound_length = 3;
    int8_t inbound_quantity = 5;
    int8_t outbound_quantity = 5;
    int8_t inbound_lengthVariance = 0;
    int8_t outbound_lengthVariance = 0;
    int8_t keepalive = 0;
    uint16_t b33UserCount = 1;
    int maxStreamCount = 2048;

    for (const auto& key: req.url_params.keys())
    {
        if (key == "name" and not constructor.setName(req.url_params.get("name")))
        {
            return { constructor, false };
        }
        else if (key == "type" and not constructor.setTunnelType( TunnelConstructor::stringToTunnelType(req.url_params.get("type")) ))
        {
            return { constructor, false };
        }
        else if (key == "inbound.length")
        {
            try { inbound_length = std::stoi(req.url_params.get("inbound.length")); } catch (...) { }
        }
        else if (key == "outbound.length")
        {
            try { outbound_length = std::stoi(req.url_params.get("outbound.length")); } catch (...) { }
        }
        else if (key == "inbound.quntity")
        {
            try { inbound_quantity = std::stoi(req.url_params.get("inbound.quntity")); } catch (...) { }
        }
        else if (key == "outbound.quntity")
        {
            try { outbound_quantity = std::stoi(req.url_params.get("outbound.quntity")); } catch (...) { }
        }
        else if (key == "inbound.lengthVariance")
        {
            try { inbound_lengthVariance = std::stoi(req.url_params.get("inbound.lengthVariance")); } catch (...) { }
        }
        else if (key == "outbound.lengthVariance")
        {
            try { outbound_lengthVariance = std::stoi(req.url_params.get("outbound.lengthVariance")); } catch (...) { }
        }
        else if (key == "b33")
        {
            try {  constructor.setBlindedLeaseSet( TunnelConstructor::stringToBlindedType(req.url_params.get("b33"))); } catch (...) { }
        }
        else if (key == "b33_user_count")
        {
            try {  b33UserCount = std::stoi(req.url_params.get("b33_user_count")); } catch (...) { }
        }
        else if (key == "transient" and not constructor.setTransient(std::string(req.url_params.get("transient")) == "true"))
        {
            return { constructor, false };
        }
        else if (key == "comments")
        {
            comments = std::string(req.url_params.get("comments")) == "true";
        }
        else if (key == "keepalive")
        {
            try { keepalive = std::stoi(req.url_params.get("keepalive")); } catch (...) { }
        }
        else if (key == "hostoverride")
        {
            try { constructor.setHostOverride( std::string(req.url_params.get("hostoverride")) == "true" ); } catch (...) { }
        }
        else if (key == "webircpassword")
        {
            try { constructor.setIrcWebPassword( std::string(req.url_params.get("webircpassword")) == "true" ); } catch (...) { }
        }
        else if (key == "ssl")
        {
            try { constructor.setSsl( std::string(req.url_params.get("ssl")) == "true" ); } catch (...) { }
        }
        else if (key == "ssl")
        {
            try { constructor.setSsl( std::string(req.url_params.get("ssl")) == "true" ); } catch (...) { }
        }
        else if (key == "max_stream_count")
        {
            try { maxStreamCount = std::stoi(req.url_params.get("max_stream_count")); } catch (...) { }
        }
        else if (key == "outproxy")
        {
            try { constructor.setProxyOutproxy( std::string(req.url_params.get("outproxy")) == "true" ); } catch (...) { }
        }
    }

    if (not constructor.setComments(comments))
    {
        return { constructor, false };
    }

    if (not constructor.setLength(inbound_length, outbound_length))
    {
        return { constructor, false };
    }

    if (not constructor.setQuantity(inbound_quantity, outbound_quantity))
    {
        return { constructor, false };
    }

    if (not constructor.setLengthVariance(inbound_lengthVariance, outbound_lengthVariance))
    {
        return { constructor, false };
    }

    if (not constructor.setKeepAlive(keepalive))
    {
        return { constructor, false };
    }

    if (not constructor.setB33UserCount(b33UserCount))
    {
        return { constructor, false };
    }

    if (not constructor.setMaxStreamCount(maxStreamCount))
    {
        return { constructor, false };
    }

    return { constructor, true };
}

std::string HttpServer::fullUrl(const crow::request &req)
{
    std::string q = req.url;
    bool first = true;
    for (const auto& key: req.url_params.keys())
    {
        if (key == "lang") continue;

        const auto val = req.url_params.get(key);
        if (! val) continue;

        if (first)
        {
            q += "?";
            first = false;
        }
        else
        {
            q += "&";
        }

        q += key + "=" + val;
    }
    return q;
}

std::string HttpServer::config(const std::string& generated, Notepad::Lang lang, bool comments, TunnelConstructor::BlindedType extraPrivacy, TunnelType type, const std::string& currentUrl)
{
    crow::mustache::context ctx;
    ctx["lang_code"] = Notepad::langToCode(lang);
    ctx["tagline"] = reinterpret_cast<const char*>(Notepad::WebUi::tagline(lang));
    ctx["payload"] = generated;
    ctx["go_back"] = reinterpret_cast<const char*>(Notepad::WebUi::goBack(lang));
    ctx["to_start_title"] = reinterpret_cast<const char*>(Notepad::WebUi::toStartTitle(lang));
    ctx["to_start"] = reinterpret_cast<const char*>(Notepad::WebUi::configToStart(lang));
    ctx["url"] = currentUrl;
    ctx["back_url"] = "/main?type=" + TunnelConstructor::tunnelTypeToString(type) + "&b33=" + TunnelConstructor::blindedTypeToString(extraPrivacy);
    ctx["version"] = VERSION;

    static const auto templ = crow::mustache::compile(HTML_CONFIG);
    std::string pageWithoutCommentsHighlight = templ.render_string(ctx);

    if (! comments) return pageWithoutCommentsHighlight;

    std::istringstream stream(pageWithoutCommentsHighlight);
    std::string buffer;
    std::string result;

    while (std::getline(stream, buffer))
    {
        if (!buffer.empty() && buffer[0] == '#')
        {
            buffer = "<span class=\"code-block-comment\">" + buffer + "</span>";
        }
        result += buffer + "\n";
    }

    return result;
}

std::string HttpServer::error(const std::string &text, Notepad::Lang lang, const std::string& backUrl, const std::string& currentUrl)
{
    crow::mustache::context ctx;
    ctx["lang_code"] = Notepad::langToCode(lang);
    ctx["tagline"] = reinterpret_cast<const char*>(Notepad::WebUi::tagline(lang));
    ctx["text"] = text;
    ctx["go_back"] = reinterpret_cast<const char*>(Notepad::WebUi::goBack(lang));
    ctx["back_url"] = backUrl;
    ctx["url"] = currentUrl;
    ctx["version"] = VERSION;

    static const auto templ = crow::mustache::compile(HTML_ERROR);
    return templ.render_string(ctx);
}

std::string HttpServer::mainServer(Notepad::Lang lang, TunnelType type, TunnelConstructor::BlindedType blinded)
{
    std::string templateString = HTML_MAIN;
    constexpr char FORM_PLACEHOLDER[] = "{{form}}";
    auto posOfPlaceholder = templateString.find(FORM_PLACEHOLDER);
    if (posOfPlaceholder == std::string::npos)
    {
        return "std::string HttpServer::mainServer internal critical error (no form placeholder in template)";
    }

    crow::mustache::context ctx;

    if (blinded == TunnelConstructor::BlindedType::keys)
    {
        templateString.replace(posOfPlaceholder, sizeof(FORM_PLACEHOLDER), HTML__FORM_SERVER_B33_WITH_AUTH);
        ctx["b33_description"] = reinterpret_cast<const char*>(Notepad::WebUi::mainB33KeysServer(lang));
    }
    else if (blinded == TunnelConstructor::BlindedType::password)
    {
        templateString.replace(posOfPlaceholder, sizeof(FORM_PLACEHOLDER), HTML__FORM_SERVER_B33_WITH_AUTH);
        ctx["b33_description"] = reinterpret_cast<const char*>(Notepad::WebUi::mainB33PasswordServer(lang));
    }
    else
    {
        templateString.replace(posOfPlaceholder, sizeof(FORM_PLACEHOLDER), HTML__FORM_SERVER_WITHOUT_B33);
    }

    constexpr char SPECIFIC_ROW_PLACEHOLDER[] = "{{specific_row}}";
    posOfPlaceholder = templateString.find(SPECIFIC_ROW_PLACEHOLDER);
    if (type == TunnelType::HttpServer)
    {
        if (posOfPlaceholder == std::string::npos)
        {
            return "std::string HttpServer::mainServer internal critical error (no specific_row placeholder in template)";
        }

        templateString.replace(posOfPlaceholder, sizeof(SPECIFIC_ROW_PLACEHOLDER), HTML__TD_HTTP_SERVER_HOSTOVERRIDE);
        ctx["hostoverride"] = reinterpret_cast<const char*>(Notepad::WebUi::mainHostoverride(lang));
    }
    else if (type == TunnelType::IrcServer)
    {
        if (posOfPlaceholder == std::string::npos)
        {
            return "std::string HttpServer::mainServer internal critical error (no specific_row placeholder in template)";
        }

        templateString.replace(posOfPlaceholder, sizeof(SPECIFIC_ROW_PLACEHOLDER), HTML__TD_IRC_SERVER_WEBPASSWORD);
        ctx["webircpassword"] = reinterpret_cast<const char*>(Notepad::WebUi::mainWebIrcPasswordServer(lang));
    }
    else
    {
        templateString.replace(posOfPlaceholder, sizeof(SPECIFIC_ROW_PLACEHOLDER), "");
    }

    ctx["lang_code"] = Notepad::langToCode(lang);
    ctx["tagline"] = reinterpret_cast<const char*>(Notepad::WebUi::tunnelType(lang, type));
    ctx["type"] = TunnelConstructor::tunnelTypeToString(type);
    ctx["blinded_type"] = TunnelConstructor::blindedTypeToString(blinded);
    ctx["to_start_title"] = reinterpret_cast<const char*>(Notepad::WebUi::toStartTitle(lang));
    ctx["go_back"] = reinterpret_cast<const char*>(Notepad::WebUi::goBack(lang));
    ctx["th_option"] = reinterpret_cast<const char*>(Notepad::WebUi::mainThOption(lang));
    ctx["th_value"] = reinterpret_cast<const char*>(Notepad::WebUi::mainThInput(lang));
    ctx["tunnel_name"] = reinterpret_cast<const char*>(Notepad::WebUi::mainTunnelName(lang));
    ctx["tunnel_name_p"] = reinterpret_cast<const char*>(Notepad::WebUi::mainTunnelName(lang));
    ctx["inbound"] = reinterpret_cast<const char*>(Notepad::WebUi::mainInbound(lang));
    ctx["outbound"] = reinterpret_cast<const char*>(Notepad::WebUi::mainOutbound(lang));
    ctx["length"] = reinterpret_cast<const char*>(Notepad::WebUi::mainLength(lang));
    ctx["quantity"] = reinterpret_cast<const char*>(Notepad::WebUi::mainQuantity(lang));
    ctx["variance"] = reinterpret_cast<const char*>(Notepad::WebUi::mainVariance(lang));
    ctx["transient"] = reinterpret_cast<const char*>(Notepad::WebUi::mainTransient(lang));
    ctx["ssl"] = reinterpret_cast<const char*>(Notepad::WebUi::mainSsl(lang));
    ctx["comments"] = reinterpret_cast<const char*>(Notepad::WebUi::mainComments(lang));
    ctx["generate"] = reinterpret_cast<const char*>(Notepad::WebUi::mainGenerate(lang));
    ctx["url"] = "/main?type=" + TunnelConstructor::tunnelTypeToString(type) + "&" + "b33=" + TunnelConstructor::blindedTypeToString(blinded);
    ctx["version"] = VERSION;

    return crow::mustache::compile(templateString).render_string(ctx);
}

std::string HttpServer::mainClient(Notepad::Lang lang, TunnelType type)
{
    static std::string clientTemplateOriginal;

    if (clientTemplateOriginal.empty())
    {
        std::string templateString = HTML_MAIN;
        const char FORM_PLACEHOLDER[] = "{{form}}";
        auto posOfPlaceholder = templateString.find(FORM_PLACEHOLDER);
        if (posOfPlaceholder == std::string::npos)
        {
            return "std::string HttpServer::mainClient internal critical error (no form placeholder in template)";
        }
        templateString.replace(posOfPlaceholder, sizeof(FORM_PLACEHOLDER), HTML__FORM_CLIENT);
        clientTemplateOriginal = templateString;
    }

    std::string clientTemplate = clientTemplateOriginal;

    crow::mustache::context ctx;

    std::string specificPart;

    if (type == TunnelType::HttpProxy or type == TunnelType::SocksProxy)
    {
        specificPart = HTML__TD_OUTPROXY;
        ctx["outproxy"] = reinterpret_cast<const char*>(type == TunnelType::HttpProxy ? Notepad::WebUi::mainHttpOutproxyClient(lang) : Notepad::WebUi::mainSocksOutproxyClient(lang));
    }

    if (type != TunnelType::SocksProxy)
    {
        specificPart += HTML__TD_CLIENT_B33;
        ctx["b33_with_auth"] = reinterpret_cast<const char*>(Notepad::WebUi::mainClientB33(lang));
    }

    constexpr char SPECIFIC_ROW_PLACEHOLDER[] = "{{specific_row}}";
    const auto posOfPlaceholder = clientTemplate.find(SPECIFIC_ROW_PLACEHOLDER);
    if (posOfPlaceholder == std::string::npos)
    {
        return "std::string HttpServer::mainClient internal critical error (no specific_row placeholder in template)";
    }
    clientTemplate.replace(posOfPlaceholder, sizeof(SPECIFIC_ROW_PLACEHOLDER), specificPart);

    ctx["lang_code"] = Notepad::langToCode(lang);
    ctx["tagline"] = reinterpret_cast<const char*>(Notepad::WebUi::tunnelType(lang, type));
    ctx["type"] = TunnelConstructor::tunnelTypeToString(type);
    ctx["to_start_title"] = reinterpret_cast<const char*>(Notepad::WebUi::toStartTitle(lang));
    ctx["go_back"] = reinterpret_cast<const char*>(Notepad::WebUi::goBack(lang));
    ctx["th_option"] = reinterpret_cast<const char*>(Notepad::WebUi::mainThOption(lang));
    ctx["th_value"] = reinterpret_cast<const char*>(Notepad::WebUi::mainThInput(lang));
    ctx["tunnel_name"] = reinterpret_cast<const char*>(Notepad::WebUi::mainTunnelName(lang));
    ctx["tunnel_name_p"] = reinterpret_cast<const char*>(Notepad::WebUi::mainTunnelName(lang));
    ctx["inbound"] = reinterpret_cast<const char*>(Notepad::WebUi::mainInbound(lang));
    ctx["outbound"] = reinterpret_cast<const char*>(Notepad::WebUi::mainOutbound(lang));
    ctx["length"] = reinterpret_cast<const char*>(Notepad::WebUi::mainLength(lang));
    ctx["quantity"] = reinterpret_cast<const char*>(Notepad::WebUi::mainQuantity(lang));
    ctx["variance"] = reinterpret_cast<const char*>(Notepad::WebUi::mainVariance(lang));
    ctx["transient"] = reinterpret_cast<const char*>(Notepad::WebUi::mainTransient(lang));
    ctx["keepalive"] = reinterpret_cast<const char*>(Notepad::WebUi::mainKeepalive(lang));
    ctx["ssl"] = reinterpret_cast<const char*>(Notepad::WebUi::mainSsl(lang));
    ctx["comments"] = reinterpret_cast<const char*>(Notepad::WebUi::mainComments(lang));
    ctx["generate"] = reinterpret_cast<const char*>(Notepad::WebUi::mainGenerate(lang));
    ctx["url"] = "/main?type=" + TunnelConstructor::tunnelTypeToString(type);
    ctx["version"] = VERSION;

    return crow::mustache::compile(clientTemplate).render_string(ctx);
}
